#include <linux/module.h>
#include <linux/fs.h>
#include <linux/device.h>
#include <linux/platform_device.h>
#include <linux/proc_fs.h>
#include <linux/sched.h>
#include <asm/uaccess.h>

static char internal_buffer[100];
static int pos = 0;
static wait_queue_head_t wq;
DECLARE_MUTEX( s );

struct buf_stat {
	atomic_t open;
	atomic_t close;
	atomic_t read;
	atomic_t write;
};
static struct buf_stat stat;
static struct proc_dir_entry *proc_dir, *proc_access;

static ssize_t stat_access( struct device_driver *driver, char *buffer )
{
	sprintf( buffer, "open: %d\nclose: %d\nread: %d\nwrite: %d\n",
		atomic_read( &stat.open ),
		atomic_read( &stat.close ),
		atomic_read( &stat.read ),
		atomic_read( &stat.write ) );
	return strlen(buffer)+1;
}
static ssize_t stat_reset( struct device_driver *driver, const char *buf,
	size_t count )
{
	atomic_set( &stat.open, 0 );
	atomic_set( &stat.close, 0 );
	atomic_set( &stat.read, 0 );
	atomic_set( &stat.write, 0 );
	return count;
}

static int proc_read( char *buf, char **start, off_t offset,
	int size, int *eof, void *data )
{
	sprintf( buf, "open: %d\nclose: %d\nread: %d\nwrite: %d\n",
		atomic_read( &stat.open ),
		atomic_read( &stat.close ),
		atomic_read( &stat.read ),
		atomic_read( &stat.write ) );
	*eof = 1;
	return strlen(buf)+1;
}

static DRIVER_ATTR( access, S_IRUGO, stat_access, NULL );
static DRIVER_ATTR( reset, S_IWUGO, NULL, stat_reset );

static int driver_open( struct inode *geraetedatei, struct file *instanz )
{
	printk("driver_open\n");
	atomic_inc( &stat.open );
	return 0;
}

static int driver_release( struct inode *geraetedatei, struct file *instanz )
{
	printk("driver_release\n");
	atomic_inc( &stat.close );
	return 0;
}

static ssize_t driver_read(struct file *instanz, char __user *buf,
	size_t count, loff_t *offset )
{
	int tocopy=0, not_copied=0;

	atomic_inc( &stat.read );
	down_interruptible( &s );
	while( pos==0 ) {
		printk("going to sleep ...\n");
		up( &s );
		if( wait_event_interruptible(wq,(pos>0)) )
			return -ERESTART;
		down_interruptible( &s );
	}
	printk("woke up ...\n");
	tocopy=min(pos,(int)count);
	not_copied=copy_to_user(buf,internal_buffer+pos-tocopy,tocopy);
	pos = pos-tocopy;
	up( &s );
	return tocopy-not_copied;
}

static ssize_t driver_write(struct file *instanz,const char __user *buf,
	size_t count, loff_t *offset )
{
	int maxleft, tocopy, not_copied;

	atomic_inc( &stat.write );
	printk("driver_write...\n");
	down_interruptible( &s );
	maxleft = sizeof(internal_buffer)-pos;
	tocopy  = min( maxleft, (int)count );
	not_copied = copy_from_user( internal_buffer+pos, buf, tocopy );
	pos += tocopy-not_copied;
	up( &s );
	wake_up( &wq );
	return tocopy-not_copied;
}

static struct file_operations fops = {
	.owner = THIS_MODULE,
	.open  = driver_open,
	.release = driver_release,
	.read = driver_read,
	.write = driver_write,
};

static struct device_driver buf_driver = {
	.name = "buf-driver",
	.bus  = &platform_bus_type,
};

static int __init sysproc_init(void)
{
	printk("sysproc_init called\n");
	if( register_chrdev( 240, "buf", &fops )!=0 )
		goto failed;
	if( driver_register( &buf_driver )!= 0 )
		goto unreg;
	init_waitqueue_head( &wq );
	atomic_set( &stat.open, 0 );
	atomic_set( &stat.close, 0 );
	atomic_set( &stat.read, 0 );
	atomic_set( &stat.write, 0 );
	if( driver_create_file( &buf_driver, &driver_attr_access ) )
		goto dr_unreg;
	if( driver_create_file( &buf_driver, &driver_attr_reset ) )
		goto drcr_unreg;
	proc_dir = proc_mkdir( "access-stat", NULL );
	proc_access = create_proc_read_entry( "access", S_IRUGO,
		proc_dir, proc_read, NULL );
	return 0;
drcr_unreg:
	driver_remove_file( &buf_driver, &driver_attr_access );
dr_unreg:
	driver_unregister( &buf_driver );
unreg:
	unregister_chrdev( 240, "buf" );
failed:
	printk("loading failed\n");
	return -EIO;
}

static void __exit sysproc_exit(void)
{
	if( proc_access ) remove_proc_entry( "access", proc_dir );
	if( proc_dir ) remove_proc_entry( "access-stat", NULL );
	driver_remove_file( &buf_driver, &driver_attr_access );
	driver_remove_file( &buf_driver, &driver_attr_reset );
	driver_unregister( &buf_driver );
	unregister_chrdev( 240, "buf" );
	printk("sysproc_exit called\n");
}

module_init( sysproc_init );
module_exit( sysproc_exit );
MODULE_LICENSE("GPL");
